#
# This awk script reads an OFF file and produces a Postscript plot;
# it plots the triangle itself and associated lat/lon grid;
# it invokes the "worldmap" program to generate geographic data.
#
function deg(angle, sig, fmt) {
    sign = substr(sig, 1, 1);
    if(angle < 0) { angle = -angle; sign = substr(sig,2,1); }
    return sprintf(fmt "%s", angle, sign);
}
function gridfmt( incr ) {
    if(incr == int(incr)) {
	return "%d";
    }
    i10 = incr*10 + .01
    if(i10 - int(i10) < .015) {
	return "%.1f";
    } else {
	return "%.2f";
    }
}
function dm(angle, sig) {
    sign = substr(sig, 1, 1);
    if(angle < 0) { angle = -angle; sign = substr(sig,2,1); }
    degr = angle;
    return sprintf("%3d:%04.1f%s", int(degr), 60*(degr - int(degr)), sign);
}
function project(x, y, z) {
    xp = x*x2 + y*y2 + z*z2;
    yp = x*x1 + y*y1 + z*z1;
    zp = x*x0 + y*y0 + z*z0;
    xproj = pscale*xp/(1+zp);
    yproj = pscale*yp/(1+zp);
#XXX
    ss = sprintf("%.1f %.1f ", xproj+xbase, yproj+ybase);
    print "%",x,y,z,"=>",xp,yp,zp,"@",pscale,"=>",ss
    return sprintf("%.1f %.1f ", xproj+xbase, yproj+ybase);
}
# Is this point inside the safe area?
# Safe area has a T-shape, leaving room for labels at right-side corners:
#   ----	<- ypmax0
#  |    |
#  |    ----	<- ypmax1
#  |        |
#  |    ----	<- ypmin1
#  |    |
#   ----	<- ypmin0
#  ^xpmin    ^xpmax
#       ^xpcut
function inside(lat, lon) {
    projectll(lat, lon);
    return xproj > xpmin && xproj < xpmax && \
	(yproj > ypmin1 || (yproj > ypmin0 && xproj < xpcut)) && \
	(yproj < ypmax1 || (yproj < ypmax0 && xproj < xpcut));
}
function edgeeqn(ind, slop) {
    p1 = $((ind+1)%3 + 2);
    p2 = $(ind%3 + 2);
    project(x[p1],y[p1],z[p1]);
    dxp = xproj; dyp = yproj;
    project(x[p2],y[p2],z[p2]);
    dxp -= xproj; dyp -= yproj;
    rproj = sqrt(dxp*dxp + dyp*dyp); dxp /= rproj; dyp /= rproj;
    dcp = -(dyp*xproj - dxp*yproj);
    # Choose sign so that (equation)[0,0] > 0, since mid-face = [0,0].
    if(dcp < 0) { dcp = -dcp;  dxp = -dxp;  dyp = -dyp; }
    return dyp " " (-dxp) " " dcp;  # or... dcp*(1 + slop);
}
function projectll(lat, lon) {
    rll = cos(lat / 57.29577951308);
    zll = sin(lat / 57.29577951308);
    xll = rll*cos(lon / 57.29577951308);
    yll = rll*sin(lon / 57.29577951308);
    return project(xll, yll, zll);
}
function midpoint(p1, p2) {
    xmid = x[$(p1+1)]+x[$(p2+1)];
    ymid = y[$(p1+1)]+y[$(p2+1)];
    zmid = z[$(p1+1)]+z[$(p2+1)];
    rmid = sqrt(xmid*xmid + ymid*ymid + zmid*zmid);
    return project(xmid/rmid, ymid/rmid, zmid/rmid);
}

function usefont(name, size) {
    htext = int(size * 6/boxpfrac);
    fontstr = name " " htext;
    outlinewidth = htext * .2;
}

function at(xpos, ypos) {
    xtext = xbase + xpos;
    ytext = ybase + ypos;
}
function nl() { 
    return xtext " " (ytext -= htext*1.5);
}
function labelll(lat, lon, lab, rot, adj, vadj) {
    if(lat >= 90 || lat <= -90) return;
    llpos = projectll(lat, lon);
    # Use xll,yll,zll generated by projectll to find orientation
    lyo = -xll*x1 - yll*y1 + (1-zll)*z1;
    lxo = -xll*x2 - yll*y2 + (1-zll)*z2;
    if(lxo == 0 && lyo == 0) lrot = rot;
    else lrot = rot+atan2(lyo,lxo)*57.29577951308;
    # if(lrot < -90 || lrot > 90) { lrot += 180; adj = -adj; }
    text(llpos, lab, lrot, adj, vadj);
}

function round(v, unit) {
    return int(v/unit)*unit;
}

####  Postscript-specific routines.

function moveto(x,y) { print x+xbase,y+ybase, "m"; }
function lineto(x,y) { print x+xbase,y+ybase, "l"; }
function stroke() { print "S"; }
function width(w) { print w,"w"; }
function setgray(g) { print g,"G"; }
function comment(str) { print "% ", str; }
function clip() { print "clip"; }
function beginpage(name) {
    if(begundoc == 0) begindoc();
    print "%%Page: " name, begundoc;
    print "/Helvetica 12 0 0 0 z";
    print "gsave";
}

function endpage(eof) {
    print "grestore";
    if(xstep > 0) {
	delta = bx1 - bx0 + xstep;
	if(eof || ((begundoc += delta) >= xstrip)) {
	    print "%%PageTrailer";
	    print "showpage";
	    print "%%Trailer";

	    if(turn > 0) {
		printf BBoxFmt, by0,bx0, by1,bx1+begundoc-delta;
	    } else {
		printf BBoxFmt, bx0,by0, bx1+begundoc-delta,by1;
	    }
	    begundoc = 0;
	} else {
	    print delta " 0 translate";
	}
    } else {
	print "%%PageTrailer";
	print "showpage";
	print "%%Trailer";
    }
}

function text(posn, lab, rot, adj, vadj) {
    srot = sin(rot/57.29577951308); crot = cos(rot/57.29577951308);
    print "/" fontstr " 0 0 " 1-adj " z";
    if(outlinewidth != oldoutlinewidth) {
	print "/outwidth " outlinewidth " def";
	oldoutlinewidth = outlinewidth;
    }
    if(vadj+0 != 0) {
	split(posn, tposn);
	posn = (tposn[1] - vadj*htext/2*srot) " " (tposn[2] + vadj*htext/2*crot);
    }
    printf "[%.3f %.3f %.3f %.3f %s]e\n", crot, srot, -srot, crot, posn;
    print lab "t";
    print "T";
}

function begindoc() {
    print "%!PS-Adobe-2.0";
    print "%%Creator:Adobe Illustrator 88(TM)";
    print "%%Real-Creator: plottri";
    xpect = 0;
    if(xstep > 0) xpect = (bx1-bx0 + xstep) * int(xstrip / (bx1-bx0 + xstep));
    BBoxFmt = "%%%%BoundingBox:%d %d %d %d\n";
    if(turn > 0) {
	printf BBoxFmt, by0,bx0, by1,bx1+xpect;
    } else {
	printf BBoxFmt, bx0,by0, bx1+xpect,by1;
    }
    print "%%EndComments";
    print "% Postscript scale = 1/" pfrac " of actual size.";
    print "%%BeginProcSet:plottri";
    if(turn > 0) print by1+bx0, by0, "translate 90 rotate", -bx0, -by0, "translate";
    print "/Courier findfont 12 scalefont setfont";
    print "/adjshow {      % str adj -- .    adj 1 = left-just, -1 = right-just";
    print "	gsave";
    print "	/aadj exch 1 sub .5 mul def";
    print "	dup stringwidth pop aadj mul 0 rmoveto";
    print "	gsave 1 setgray outwidth setlinewidth dup false charpath stroke";
    print "	grestore show";
    print "	grestore";
    print "} def";
    print "/m {moveto} def  /l {lineto} def  /S {stroke} def /s {closepath S}def";
    print "/G {setgray} def /g {setgray} def /j {setlinejoin} def";
    print "/J {setlinecap} def /w {setlinewidth} def  /n {newpath} def";
    print "/M {setmiterlimit} def /i {pop} def";
    print "/z { neg 1 add /zadj exch def pop pop exch findfont exch scalefont setfont } def";
    print "/t { zadj adjshow } def";
    print "/U {} def /u {} def";
    print "/e { gsave concat 0 0 moveto } def /T {grestore} def";
    print "%%EndProcset";
    print "%%EndProlog";
    print "0 G  0 i 0 J 0 j 4 M";
    width(1);
}

### End postscript

BEGIN { first = 1; }

first==1 {
    first = $0;
    if(scale=="") scale = 6397;  # cm radius
    # Actual size is "scale" inches per radian (at the stereographic pole)
    # Postscript size is "pscale/(2*72)" inches per radian
    if(pfrac=="") pfrac = 4.5;	# Actual size / Postscript size factor
    if(pscale=="") pscale = 2*72*(scale/2.54)/pfrac;
    pfrac = 2*72*scale/pscale;
    if(boxpfrac == "") boxpfrac = pfrac;
    if(precision=="") precision = 90;	# Map precision in arc seconds
    if(xstrip == "") xstrip = 13*12*72/pfrac;  # Default 13-foot strips


    while((getline < "tritypes") > 0) {
	type[$2] = $1;
    }

    xbase = 160*(6/boxpfrac);    ybase = 280*(6/boxpfrac);
    ypmin0 = -235*(6/boxpfrac);  ypmin1 = -165*(6/boxpfrac);
    ypmax0 =  235*(6/boxpfrac);  ypmax1 =  157*(6/boxpfrac);
    xpmin = -130*(6/boxpfrac);   xpmax = 250*(6/boxpfrac);   xpcut = 90*(6/boxpfrac);

    bx0 = xbase+xpmin-90/boxpfrac;   bx1 = xbase+xpmax+90/boxpfrac;
    by0 = ybase+ypmin0-90/boxpfrac;  by1 = ybase+ypmax0+90/boxpfrac;

    if(length(grid) == 0) grid = 1;
    if(longrid!="") rawlongrid = longrid;
    if(latgrid=="") latgrid = grid;

    head1 = "         Side [cm] [assuming " scale " cm radius]";
    head2 = "Type   A     B     C	  corner a	 corner b	corner c  center";
    head = head1 "\n" head2;

    begundoc = 0;

    if(sublabels != "") {
	i = 0;
	while((getline < sublabels) > 0) {
	    if($0 != "") sublbl[i++] = $0;
	}
    }
    $0 = first;
    first = 0;
}

/^[ 	]*#.*/ { next; }

/OFF/ { next; }

NF == 3 && nv == "" {
    nv = $1; nf = $2;
    vno = 0;
    fno = 0;
    next;
}
NF >= 3 && vno < nv {
    x[vno] = $1; y[vno] = $2; z[vno] = $3;
    mx += $1; my += $2; mz += $3;
    if(maxz < $3) maxz = $3;
    if(minz < $3) minz = $3;
    if($5 == "") {
	vname[vno] = "";
    } else {
	split($5, tvname, "/");
	vname[vno] = sprintf("%2s/%-2s", tvname[1], tvname[2]);
    }
    vno++;
    next;
}

NF >= 4 {
    lastadj = "";
    beginpage($6);
    moveto(bx0-xbase,by0-ybase); lineto(bx1-xbase,by0-ybase);
    lineto(bx1-xbase,by1-ybase); lineto(bx0-xbase,by1-ybase);
    lineto(bx0-xbase,by0-ybase); stroke();

    moveto(bx0-xbase,by0-ybase); lineto(bx1-xbase,by0-ybase);
    lineto(bx1-xbase,by1-ybase); lineto(bx0-xbase,by1-ybase);
    lineto(bx0-xbase,by0-ybase); clip();

    x0 = x[$4]+x[$2]+x[$3];
    y0 = y[$4]+y[$2]+y[$3];
    z0 = z[$4]+z[$2]+z[$3];

    R = sqrt(x0*x0 + y0*y0 + z0*z0); x0 /= R; y0 /= R; z0 /= R;
    r0 = sqrt(x0*x0 + y0*y0);

    midlon = 57.29577951308*atan2(y0,x0);
    midlat = 57.29577951308*atan2(z0,r0);

    nfv = $1;
    latmin = 999; latmax = -999;  lonmin = 999;  lonmax = -999;
    for(i=2; i<=4; i++) {
	j=(i-1)%3 + 2;
	dx = x[$i]-x[$j];
	dy = y[$i]-y[$j];
	dz = z[$i]-z[$j];
	r2 = dx*dx + dy*dy + dz*dz;

	l[8-i-j] = 2*scale*atan2(sqrt(r2/4), sqrt(1-r2/4));
	r = sqrt(x[$i]*x[$i] + y[$i]*y[$i]);
	dlat = atan2(z[$i],r) * 57.29577951308;
	if(latmin > dlat) latmin = dlat;
	if(latmax < dlat) latmax = dlat;
	if(r > 0) {
	    dlon = atan2(y[$i], x[$i]) * 57.29577951308;
	    if(dlon - midlon > 120) dlon -= 360;
	    else if(dlon - midlon < -120) dlon -= 360;
	    if(lonmin > dlon) lonmin = dlon;
	    if(lonmax < dlon) lonmax = dlon;
	} else {
	    dlon = 0;
	}

	lat = dm(dlat, "NS");
	lon = dm(dlon, "EW");
	pos[i-1] = lat " " lon;
    }

    for(i=1; i<=3; i++) tl[i] = l[i] + .002*z[$(i+1)];

    if(tl[1]<tl[2]) { min=3; if(tl[1]<tl[3]) min=1; }
    else if(tl[2]<tl[3]) min=2; else min=3;

    if(tl[1]>=tl[2]) { max=3; if(tl[1]>=tl[3]) max=1; }
    else if(tl[2]>=tl[3]) max=2; else max=3;
    mid = 6-max-min;

    # up vector v1 is (vertex B) - (vertex A).
    x1 = x[$(mid+1)] - x[$(min+1)];
    y1 = y[$(mid+1)] - y[$(min+1)];
    z1 = z[$(mid+1)] - z[$(min+1)];
    # Orthogonalize with stereographic projection vect.
    dot = x1*x0 + y1*y0 + z1*z0;
    x1 -= x0*dot;  y1 -= y0*dot;  z1 -= z0*dot;
    R = sqrt(x1*x1 + y1*y1 + z1*z1);  x1 /= R; y1 /= R; z1 /= R;
    
    # right vector is cross(v0, v1)
    x2 = y1*z0 - z1*y0;  y2 = z1*x0 - x1*z0;  z2 = x1*y0 - y1*x0;

    # Ensure 'C' vertex is at right; rotate 180 degrees if necessary.
    project(x[$(max+1)], y[$(max+1)], z[$(max+1)]);
    if(xproj < 0) {
	x1 = -x1; y1 = -y1; z1 = -z1;
	x2 = -x2; y2 = -y2; z2 = -z2;
    }
    project(x[$(min+1)], y[$(min+1)], z[$(min+1)]);
    Aup = 1.5;
    if(yproj < 0) Aup = -1.5;

    comment( "%% Basis" );
    comment( sprintf("%% %.5f %.5f %.5f", x2,y2,z2) );
    comment( sprintf("%% %.5f %.5f %.5f", x1,y1,z1) );
    comment( sprintf("%% %.5f %.5f %.5f", x0,y0,z0) );

    width(2);
    project(x[$2],y[$2],z[$2]); moveto(xproj,yproj);
    project(x[$3],y[$3],z[$3]); lineto(xproj,yproj);
    project(x[$4],y[$4],z[$4]); lineto(xproj,yproj);
    project(x[$2],y[$2],z[$2]); lineto(xproj,yproj);
    stroke();
    width(.5);

    slat = dm(midlat, "NS");
    midface = slat  " " dm(midlon, "EW");

    usefont("Helvetica-Bold", 20);

    text( midpoint(mid,max), "(a)", 0, 0, -Aup-.5);
    text( midpoint(mid,mid), "(B)", 0, 0, -Aup-.5);
    text( midpoint(min,max), "(b)", 0, 0, Aup-.5);
    text( midpoint(min,mid), "( c)", 0, 1, 0);
    text( midpoint(min,min), "(A)", 0, 0, Aup-.5);
    text( midpoint(max,max), "(C)", 0, 1, -.5);

    if(grid > 0) {
      longridscale = 1;
      if(midlat < -50 || midlat > 50) {
	longridscale = 2;
	if(midlat < -70 || midlat > 70) {
	    longridscale = 5;
	    #if(midlat < -80 || midlat > 80) longrid = 10;
	}
      }
      if(rawlongrid == "") rawlongrid = grid;
      longrid = rawlongrid * longridscale;

      latmin = round(latmin-latgrid, latgrid); if(latmin < -90) latmin = -90;
      latmax = round(latmax+latgrid,latgrid); if(latmax > 90) latmax = 90;
      lonmin = round(lonmin-longrid, longrid);
      lonmax = round(lonmax+longrid, longrid);

      latfmt = gridfmt( latgrid )
      lonfmt = gridfmt( longrid )

      usefont("Helvetica", 7);
    
      for(lat = latmin; lat <= latmax; lat += latgrid) {
	slat = deg(lat, "NS", latfmt);
	step = 1;
	for(lon=lonmin; lon<=lonmax && step > .1; lon += longrid*step) {
	    if(inside(lat, lon)) { lon -= longrid*step; step /= 2; }
	}
	lon0 = lon;
	step = 1;
	for(lon=lonmax; lon>=lonmin && step > .125; lon -= longrid*step) {
	    if(inside(lat, lon)) { lon += longrid*step; step /= 2; }
	}
	lon1 = lon;
	if(lon0 > lon1) continue;
	labelll(lat, lon0, "(" slat " )", -90, 1, .5);
	labelll(lat, lon1, "( " slat ")", -90, -1, .5);
	projectll(lat, lon0); moveto(xproj,yproj);
	for(lon = round(lon0+longrid,longrid); lon < lon1; lon += longrid) {
	    projectll(lat, lon); lineto(xproj,yproj);
	}
	projectll(lat, lon1); lineto(xproj,yproj);
	stroke();
      }

      for(lon = lonmin; lon <= lonmax; lon += longrid) {
	slon = deg(lon, "EW", lonfmt);
	step = 1;
	for(lat=latmin; lat<=latmax && step > .1; lat += latgrid*step) {
	    if(inside(lat, lon)) { lat -= latgrid*step; step /= 2; }
	}
	lat0 = lat;  if(lat0 < -90) lat0 = -90;
	step = 1;
	for(lat=latmax; lat>=latmin && step > .1; lat -= latgrid*step) {
	    if(inside(lat,lon)) { lat += latgrid*step; step /= 2; }
	}
	lat1 = lat;	 if(lat1 > 90) lat1 = 90;
	if(lat0 > lat1) continue;
	labelll(lat0, lon, "(" slon ")", 0, 1, .5);
	labelll(lat1, lon, "(" slon ")", 0, -1, .5);

	projectll(lat0, lon); moveto(xproj,yproj);
	for(lat=round(lat0+latgrid,latgrid); lat < lat1; lat += latgrid) {
	    projectll(lat, lon); lineto(xproj,yproj);
	}
	projectll(lat1, lon); lineto(xproj,yproj);
	stroke();
      }
    }


    cmd = "set -x; echo -n longrid=" longrid ", latgrid=" latgrid ", \"" $6 " \" >&2";
    if(precision > 0) {
	if(noclip == "") {
	    clips = " -c " edgeeqn(min, .25);
	    clips = clips " -c " edgeeqn(mid, .25);
	    clips = clips " -c " edgeeqn(max, .15);
	}
	cmd = cmd "; worldmap -p " precision \
	    " -s " pscale \
	    " -o " xbase " " ybase \
	    clips \
	    " -P "	x2 " " y2 " " z2 " " \
		    x1 " " y1 " " z1 " " \
		    x0 " " y0 " " z0 " " \
	    latmin " " latmax " " lonmin " " lonmax;
	comment( cmd );
    }
    system(cmd);

    if(subdiv > 1) {
	ssubdiv = sqrt(subdiv);
	setgray(.5);
	for(w = 1; w <= 3; w++) {
	    k0 = $(w+1);
	    k1 = $((w%3)+2);
	    k2 = $(((w+1)%3)+2);
	    for(i = 1; i < subdiv; i++) {
		tu = i/subdiv;
		ex1 = tu*x[k1]+(1-tu)*x[k0];
		ey1 = tu*y[k1]+(1-tu)*y[k0];
		ez1 = tu*z[k1]+(1-tu)*z[k0];

		ex2 = tu*x[k2]+(1-tu)*x[k0];
		ey2 = tu*y[k2]+(1-tu)*y[k0];
		ez2 = tu*z[k2]+(1-tu)*z[k0];

		if(i % ssubdiv == 0) width(1);
		else width(.5);

		project(ex2, ey2, ez2); moveto(xproj, yproj);
		for(j = 1; j <= i; j++) {
		    tv = j/i;
		    project(tv*ex1 + (1-tv)*ex2, \
			    tv*ey1 + (1-tv)*ey2, tv*ez1 + (1-tv)*ez2);
		    lineto(xproj, yproj);
		}
		stroke();
	    }
	}
	setgray(0);
	width(.5);
    }
    if(sublabels != "") {
	w = 0;
	for(i = 0; i < subdiv; i++) {
	    tu = (i + .5) / subdiv;
	    ex1 = (1-tu)*x[$3] + tu*x[$2];
	    ey1 = (1-tu)*y[$3] + tu*y[$2];
	    ez1 = (1-tu)*z[$3] + tu*z[$2];

	    ex2 = (1-tu)*x[$3] + tu*x[$4];
	    ey2 = (1-tu)*y[$3] + tu*y[$4];
	    ez2 = (1-tu)*z[$3] + tu*z[$4];
		# interpolate from ($4~$3) to ($2~$3)
	    for(j = 0; j <= 2*i; j++) {
		tv = (j + .5) / (2*i + 1);
		text( project(tv*ex1 + (1-tv)*ex2, \
			tv*ey1 + (1-tv)*ey2, tv*ez1 + (1-tv)*ez2),
		      "(" sublbl[w++] ")",
		      0,
		      0, 0);
	    }
	}
    }

    usefont("Helvetica-Bold", 20);
    at((xpmax+xpcut)/2, ypmax0 + .25*htext);
    text( xtext " " ytext, "(" $6 ")", 0, 0, -1 );

    usefont("Courier", 9);
    at(xpmax*.1 + xpcut*.9, ypmax0 - .25*htext);
    text( xtext " " ytext, "(Vertex)", 0, 1, -1);

    at(xpmax, ypmax0 - .25*htext);
    text( xtext " " ytext, "([type " type[substr($6,2,2)] "])", 0, -1, -1);

    at(xpmax*.1 + xpcut*.9, ypmax0 - .9*htext);
    text(nl(),sprintf("(%5s A %19s)", vname[$(min+1)], pos[min]),0,1,0);
    text(nl(),sprintf("(%5s B %19s)", vname[$(mid+1)], pos[mid]),0,1,0);
    text(nl(),sprintf("(%5s C %19s)", vname[$(max+1)], pos[max]),0,1,0);
    text(nl(),sprintf("(Center  %19s)", midface),0,1,0);
    p = (l[min]*l[min] + l[max]*l[max] - l[mid]*l[mid])/(2*l[max]);
    h = sqrt(l[min]*l[min] - p*p);
    at(xpmax, ypmin1);
    nl();
    text(nl(),sprintf("(%5.2f Height   \\(in cm,   )", h), 0,-1,0);
    text(nl(),sprintf("(%5.2f Edge a   assuming  )", l[min]), 0,-1,0);
    text(nl(),sprintf("(%5.2f Edge b   %7.2f cm)", l[mid], scale), 0,-1,0);
    text(nl(),sprintf("(%5.2f Edge c   sphere.\\)  )", l[max]), 0,-1,0);
    endpage(0);
}

END {
    if(begundoc > 0) endpage(1);
}
